#!/etc/opt/openit/perl

use strict;
use warnings;

my $arg_count = @ARGV;

if ($arg_count < 3 || $arg_count > 5) {
	show_usage();
	exit;
}

my $option = $ARGV[0];
my @tokens;
my $node_type;
my $line_num;
my $node_value;
my $file_path;
my $node_path;
if ($option eq "--set" || $option eq "-s") {
	if ($arg_count != 4) {
		print "Error: Invalid number of arguments.\n";
		show_usage();
		exit; 
	}
	$file_path = $ARGV[1];
	$node_path = $ARGV[2];
	$node_value = $ARGV[3];
	@tokens = parse_conf($file_path);
	$node_type = is_node_valid($node_path);
	if ($node_type != 1) {
		print "Error: Not a valid terminal node path.\n";
		show_usage();
		exit;
	}
	$line_num = is_node_existing(@tokens,$node_path,$node_type);
	if ($line_num == 0) {
		print "Error: Node path $node_path does not exist.\n";
		exit;
	}
	$tokens[$line_num] =~ s/\=\".*\"/\=\"$node_value\"/;
	save_file(@tokens,$file_path);
}
elsif ($option eq "--delete" || $option eq "-d") {
	if ($arg_count != 3) {
		print "Error: Invalid number of arguments.";
		show_usage();
		exit;
	}
	$file_path = $ARGV[1];
	$node_path = $ARGV[2];
	@tokens = parse_conf($file_path);
	$node_type = is_node_valid($node_path);
	if ($node_type == 0) {
		print "Error: Invalid node path.";
		show_usage();
		exit;
	}
	$line_num = is_node_existing(@tokens,$node_path,$node_type);
	if ($line_num == 0) {
		print "Error: Node path $node_path does not exist.\n";
		exit;
	}
	@tokens = delete_node(@tokens,$line_num,$node_type);
	save_file(@tokens,$file_path);
}
elsif ($option eq "--insert" || $option eq "-i") {
	print "Insert option here.\n";
}
elsif ($option eq "--copy" || $option eq "-c") {
	if ($arg_count != 5) {
		print "Error: Invalid number of arguments.\n";
		show_usage();
		exit;
	}
	my $source_path = $ARGV[1];
	my $source_node = $ARGV[2];
	my $target_path = $ARGV[3];
	my $target_node = $ARGV[4];
	if (check_source_target_nodes($source_node,$target_node) == 0) {
		print "Error: Target node $target_node must be a parent node to source node $source_node\n";
		show_usage();
		exit;
	}
	my @source_tokens = parse_conf($source_path);
	my @target_tokens = parse_conf($target_path);
	my $source_node_type = is_node_valid($source_node);
	if ($source_node_type == 0) {
		print "Error: Invalid node path $source_node\n";
		show_usage();
		exit;
	}
	my $target_node_type = is_node_valid($target_node);
	if ($target_node_type != 2) {
		print "Error: Target node path $target_node must be a parent node.\n";
		show_usage();
		exit;
	}
	my $source_line = is_node_existing(@source_tokens,$source_node,$source_node_type);
	if ($source_line == 0) {
		print "Error: Node path $source_node does not exist.\n";
		exit;
	}
	my $target_line = is_node_existing(@target_tokens,$target_node,$target_node_type);
	if ($target_line == 0) {
		print "Error: Node path $target_node does not exist.\n";
		exit;
	}
	my $node_already_exists = is_node_existing(@target_tokens,$source_node,$source_node_type);
	if ($node_already_exists) {
		print "Verbose: Node path $source_node already exist in $target_path. Aborting copy.\n";
		exit;
	}
	my @new_node = get_node_object(@source_tokens,$source_line,$source_node_type);
	if ($target_tokens[$target_line] =~ /.+({})$/) {
		$target_tokens[$target_line] =~ s/\=\"\"{}/\=\"\"/;
		my $tab_count = $target_tokens[$target_line] =~ tr/\t//;
		my $temp = substr($target_tokens[$target_line],0,$tab_count);
		$temp .= "{";
		unshift(@new_node,$temp);
		$temp =~ s/\{$/\}/;
		push(@new_node,$temp);
		splice(@target_tokens,$target_line+1,0,@new_node);
	}
	else {
		splice(@target_tokens,$target_line+2,0,@new_node);
	}
	save_file(@target_tokens,$target_path);
}
else {
	show_usage();
}

sub get_node_object {
	my (@args) = @_;
	my $type = pop(@args);
	my $line_number = pop(@args);
	my @node_block;
	my $opening_tag;
	if ($type == 1 || $args[$line_number+1] !~ /\t{$/) {
		push(@node_block,$args[$line_number]);
	}
	else {
		$opening_tag = $args[$line_number+1];
		for my $i ($line_number .. $#args) {
			push(@node_block,$args[$i]);
			if ($args[$i] =~ /\t}$/) {
				my $temp1 = $opening_tag;
				my $temp2 = $args[$i];
				chop($temp1);
				chop($temp2);
				if ($temp1 eq $temp2) {
					last;
				}
			}	
		}	
	}
	return @node_block;
}

sub check_source_target_nodes {
	my (@args) = @_;
	my @source_nodes = split(/\./,$args[0]);
	my @target_nodes = split(/\./,$args[1]);
	if ($#source_nodes != $#target_nodes+1) {
		print "$#source_nodes $#target_nodes\n";
		return 0;
	}
	for my $i (0 .. $#target_nodes) {
		if ($source_nodes[$i] ne $target_nodes[$i]) {
			return 0;
		}
	}
	return 1;
}

sub delete_node {
	my (@args) = @_;
	my $type = pop(@args);
	my $line_number = pop(@args);
	my $opening_tag;
	my $closing_tag;
	my $end_of_block;
	my $length;
	if ($type == 1) { # aterminal node
		if ($args[$line_number-1] =~ /\t{$/ && $args[$line_number+1] =~ /\t}$/) {
			$args[$line_number-2] .= "{}";
			splice(@args,$line_number-1,3);
		}
		else {
			splice(@args,$line_number,1);
		}
	}
	elsif ($type == 2) {
		if ($args[$line_number] =~ /.+\=\"\"({})$/) { # note terminal but has no children
			if ($args[$line_number-1] =~ /\t{$/ && $args[$line_number+1] =~ /\t}$/) {
				$args[$line_number-2] .= "{}";
				splice(@args,$line_number-1,3);
			}
			else {
				splice(@args,$line_number,1);
			}
		}
		else { # has child nodes
			$opening_tag = $args[$line_number+1];
			for my $i ($line_number+2 .. $#args) {
				if ($args[$i] =~ /\t}$/) {
					my $temp1 = $opening_tag;
					my $temp2 = $args[$i];
					chop($temp1);
					chop($temp2);
					if ($temp1 eq $temp2) {
						$end_of_block = $i;
						last;
					}
				}
			}
			$length = ($end_of_block-$line_number)+1;
			# check if sole child node
			if ($args[$line_number-1] =~ /\t{$/ && $args[$end_of_block+1] =~ /\t}$/) {
				$args[$line_number-2] .= "{}";
				splice(@args,$line_number-1,$length+2);
			}
			else {
				splice(@args,$line_number,$length);
			}
		}
	}
	return @args;
}

sub save_file {
	my (@args) = @_;
	my $file_path = pop(@args);

	open my $conf_file, ">", "$file_path" or die "Error: Couldn't open file $file_path: $!\n";
	for my $i (@args) {
		print $conf_file "$i\n";
	}
	close $conf_file;
}

sub is_node_existing {
	my (@toks) = @_;
	my $type = pop(@toks);
	my $path = pop(@toks);
	my $line_number = 0;
	my $node_number = 0;
	my $node_name;
	my $token;
	my @path_toks = split(/\./,$path);
	my $flag1 = 0; # 0 start, 1 initial matching, 2 still in the block
	my $flag2 = 0; # found match (not the last one), mark to beging another matching

	my $opening_tag;
	for my $i (0 .. $#toks) {
		if ($node_number > $#path_toks) {
			return 0;
		}
		if ($i == 0 && $node_number == 0) { #compare first node amd first line
			$token = $toks[$i];
			$node_name = $path_toks[$i];
			$token =~ s/\(\)\=\"\"({})?//;
			if ($token ne $node_name) {
				return $i;
			}
			$node_number++;
			$flag1 = 1; # start of another node matching omsode a block
			next;
		}
		if ($toks[$i] =~ /\t{$/ || $toks[$i] eq "{") {
			if ($flag1 == 1) {
				$opening_tag = $toks[$i];
				$flag1 = 2;
				$flag2 = 0;
				next;
			}
		}
		elsif ($toks[$i] =~ /\t}$/ or $toks[$i] eq "}") {
			my $temp1 = $opening_tag;
			my $temp2 = $toks[$i];
			chop($temp1);
			chop($temp2);
			if ($temp1 eq $temp2) {
				if ($flag2 == 0) {
					return 0;
				}
			}
		}
		else {
			$token = $toks[$i];
			$node_name = $path_toks[$node_number];
			if ($node_number < $#path_toks) {
				$token =~ s/\t//g;
				$token =~ s/\(\)\=\"\"({})?//;
				if ($token eq $node_name) {
					$node_number++;
					$flag1 = 1;
					$flag2 = 1;
				}
			}
			else {
				$node_name = $path_toks[$node_number];
				$token =~ s/\t//g;
				if ($type == 1) {
					$token =~ s/\=\".*\"({})$//;
				}
				elsif ($type == 2) {
					$token =~ s/\(\)\=\"\"({})?//;
				}
				if ($token eq $node_name) {
					return $i;
				}
			}
		}
	}
	return $line_number;
}

sub show_usage {
	print "Usage: <options> <arg1> <arg2>..\n";
	print "Options:\n";
	print "   --set or -s\n";
	print "\tFormat: --set <path/to/file> <path.to.node> <value>\n";
	print "\tE.g.    --set /var/opt/openit/etc/sample.conf \"sample.node1.node2.node3(Bool)\" true\n";
	print "   --delete or -d\n";
	print "\tFormat: --delete <path/to/file> <path.to.node>\n";
	print "\tE.g.    --delete /var/opt/openit/etc/sample.conf \"sample.node1.node2(String)\"\n";
	print "   --insert or -i\n";
	print "\tFormat: --insert <path/to/file> <path.to.node> <value>\n";
	print "\tE.g.    --insert /var/opt/openit/etc/sample.conf \"sample.node1.node2.node3(String)\" \"Hello World\"\n";
	print "\tNote: Must only insert object in nodes without type (String, Bool, etc.).\n";
	print "   --copy <path/to/source/file> <path.to.node> <path/to/target/file> <path.to.parent.node>\n";
	print "\tFormat: --copy files/sample1.conf sample1.node1.node2.node3 /var/opt/openit/etc/sample.conf sample.node1.node2\n";
}

sub parse_conf {
	my (@args) = @_;

	my $conf_name = $args[0];
	open my $conf_file, "$conf_name" or die "Error: Couldn't open file $conf_name: $!";
	my @lines = <$conf_file>;
	close $conf_file;
	
	# remove new lines
	chomp(@lines);
	my @tags;
	# get the value of 2nd and last tokens, must be equal to { and } respectively
	if ($lines[1] ne "{" || $lines[$#tokens] ne "}") {
		print "Error: Invalid conf file format.\n";
		exit;
	}
	# make sure we only have one root node so set a counter
	my $line_number = 0;
	for my $i (@lines) {
		$line_number++;
		
		# Skip white space
		if ( $i =~ /^\s*$/) {
			next;
		}
		if ($i eq "{") {
			if ($line_number != 2) { 
				print "Error: There should only be one root node. See line $line_number.\n";
				exit;
			}
		}
		elsif ($i eq "}") {
			if ($line_number != ($#lines+1)) {
				print "Error: Misplaced closing tag for root node. See line $line_number.\n";
				exit;
			}
		}
		elsif ($i =~ /\t{$/) {
			push(@tags,$i);
		}
		elsif ($i =~ /\t}$/) {
			# compare number of tabs from last element
			# if @tags is empty, show error message and exit
			if (!@tags) {
				print "Error: Missing opening tag '{'. See line $line_number.\n";
				exit;
			}
			my $item = pop(@tags);
			my $temp = $i;
			chop($item);
			chop($temp);
			if ($item ne $temp) {
				print "Error: Misaligned tag. See line $line_number.\n";
				exit;
			} 
		}
		elsif ($i !~ /\t*[A-Za-z0-9:\-_]+(\(\)\=\"\"({})?)$/ && $i !~ /\t*[A-Za-z0-9:\-_]+\((String|Bool|FileName|Real|DirName|TimeSpan|Integer|Date)\)\=\".*\"({})$/) {
			print "Error: Invalid file format: See line $line_number.\n";
			exit;
		}
	}
	
	return @lines;
}

sub is_node_valid {
	my (@args) = @_;
	my @nodes = split(/\./,$args[0]);
	my $type = 0; # invalid
	if ($#nodes+1 < 2) {
		return $type;
	}
	my $temp = pop(@nodes);
	if ($temp =~ /^([A-Za-z0-9:\-_]+)(\((String|Bool|FileName|Real|DirName|TimeSpan|Integer|Date)\))$/) {
		$type = 1; # a terminal node
	}
	else {
		$type = 2; # an open node
	}
	for my $i (@nodes) {
		if ($i !~ /^[A-Za-z0-9:\-_]+$/) {
			$type = 0;
			last;
		}
	}
	return $type;
}



