#!/etc/opt/openit/perl

BEGIN {
    push(@INC,"/opt/openit/lib" );
}

use strict;
use warnings;
use Getopt::Long;
use File::Copy qw(move);
use File::Basename;
use OpeniT::Config;
use OpeniT::DataTypes;


my $installationlog;
my $opt_debug;
my $opt_verbose;
my $opt_help;


unless ( GetOptions("help" => \ $opt_help,
         "debug"           => \ $opt_debug,
         "verbose"         => \ $opt_verbose
)) {
    &print_usage;
}

my $configdir = &get_configuration_dir();
&_update_acc_types("$configdir/acc_types");
&_update_classvar();
&_update_filter();

exit 0;

sub _update_classvar()
{
	my ($filter) = "$configdir/classvar_mapping_config";
    &create_backup($filter);
    VERBOSE( "Updating $filter" );
    open(FILE, $filter) || die "Unable to read filter file '$filter'";
    my @lines = <FILE>;	

	my $ind = 0;
    ++$ind while( $lines[$ind] !~ /Autodesk.software.inventory/ );
	$ind+=1;
	$lines[$ind] ="	( License^type ),\n";
	$ind+=1;
	$lines[$ind] ="  ((Product,Product^id), (Product^id, Feature), (Feature,Serial^number),(Serial^number,Release^number), (Host,Operating^system), (Location,Installed) ))\n";
	
	_save_config( $filter, @lines );
}



sub _update_filter()
{
	my ($filter) = "$configdir/filtrator/license-autodesk-inventory.conf";
    &create_backup($filter);
    VERBOSE( "Updating $filter" );
    open(FILE, $filter) || die "Unable to read filter file '$filter'";
    my @lines = <FILE>;	
	
	_grep_and_replace(\@lines,"releasenumber=var0.*","releasenumber=var0,licensetype=var3,location=var5,installed=var6\"{}");
	_grep_and_replace(\@lines,"interval:start:'118'.*","interval:start:'118':'0':var4:'0':product_id:'0':feature:'0':serialnumber:'0':var0:'0':var3:'0':host:'0':os:'0':var5:'0':var6\"{}");
	_save_config( $filter, @lines );
}
sub _update_acc_types
{
    my ($filter) = @_;
    &create_backup($filter);
    VERBOSE( "Updating $filter" );
    open(FILE, $filter) || die "Unable to read filter file '$filter'";
    my @lines = <FILE>;

    
    if ( grep {/Autodesk/} @lines )
    {
        _remove_datatype118( \@lines,  $filter);
		return;
    }
    
    my @insert115 = 
    (
        ",\n\n (Hostgroup^license^use^v3,\n",
            "  (\n",
            "   (Count,Record^Count,x,SUM),\n",
            "   (Elapsed^time,Accumulated^time^for^group^in^hours,Hour,SUM),\n",
            "   (Max^used,Maximum^concurrent^licenses^used^by^group,x,MAX),\n",
            "   (Avg^used,Average^concurrent^licenses^used^by^group,x,ABSAVG),\n",
            "   (Min^used,Minimum^concurrent^licenses^used^by^group,x,MIN),\n",
            "   (Duration, Duration^for^group^in^hours,Hour,SUM),\n",
            "   (License^cost,License^cost^given^in^seconds^*^cost^factor,Hour,WGT,\n",
            "    (Elapsed^time,((RANGE,olap_cpu_cost))))\n",
            "  ),\n",
            "  (Product,\n",
            "   Package,\n",
            "   Feature,\n",
            "   Feature^version,\n",
            "   Featureset,\n",
            "   Featureset^version,\n",
            "   Group^type,\n",
            "   Group,\n",
            "   License^type)\n",
            "  )"
    );
    
    # Note: not sure if i will add this for now lets add the dummy
    my @insert116 = 
    (
        ",\n\n (Total^PPU^license^use^logfile,\n",
            "  (\n",
            "   (Count,Record^Count,x,SUM),\n",
            "   (Running^time,Running^time^in^hours,Hour,SUM),\n",
            "   (Avg^in^use,Average^number^of^licenses^in^use,x,ABSAVG),\n",
            "   (Max^in^use,Maximum^number^of^licenses^in^use,x,MAX),\n",
            "   (Min^in^use,Minimum^number^of^licenses^in^use,x,MIN),\n",
            "   (Elapsed^time,Elapsed^license^time^in^hours,Hour,SUM),\n",
            "   (Elastic^usage,Elastic^usage^time^in^hours,Hour,SUM),\n",
            "   (Distinct,Always^one^for^each^extracted^unique^classification,x,IDENT)\n",
            "  ),\n",
            "  (Customer^id,\n",
            "   Product,\n",
            "   Package,\n",
            "   License^type,\n",
            "   Feature,\n",
            "   Usage^type,\n",
            "   Max^available^licenses,\n",
            "   Used^licenses)\n",
            "  )"
    );
    
    my @insert117 = 
    (
        ",\n\n (Dongle^data^inventory,\n",
            "  (\n",
            "   (Count,Record^Count,x,SUM),\n",
            "   (Running^time,Accumulated^time^for^group^in^hours,Hour,SUM)\n",
            "  ),\n",
            "  (Dongle^type,\n",
            "   Serial^number,\n",
            "   Vendor,\n",
            "   Vendor^code,\n",
            "   Feature,\n",
            "   Feature^code,\n",
            "   Host^name)\n",
            "  )"
    );

    my @insert118 = 
        (
            ",\n\n (Autodesk^software^inventory,\n",
            "  (\n",
            "   (Count,Record^Count,x,SUM)\n",
            "  ),\n",
            "  (Product,\n",
            "   Product^id,\n",
            "   Feature,\n",
            "   Serial^number,\n",
            "   Release^number,\n",
            "   License^type,\n",
            "   Host,\n",
            "   Operating^system,\n",
            "   Location,\n",
            "   Installed)\n",
            "  ) \n"
    );
    
    
    
    
    my @data_types =  list_data_type_ids();
    my $no_of_dt  =  $#data_types;
    DEBUG ("Number of datatypes: $no_of_dt");
    
    $no_of_dt++;
    my @datatypeids = ( $no_of_dt .. 117 );
    my @holder =() ;
    
    foreach my $dt ( @datatypeids )
    {
        my @dummy =(
            
            ",\n\n (Placeholder^for^datatype^$dt,\n",
            "   ((open)),\n",
            "   (open))",           
        );
        if ( $dt eq 115 )
        {
            push (@holder , @insert115 );
        }   
        if ( $dt eq 116 )
        {
            push (@holder , @dummy );
        }   
        if ( $dt eq 117 )
        {
            push (@holder , @insert117 );   
        }
            
        
    }
    
    my @insert = ( "\n" );
    push (@insert, @holder );
    push (@insert, @insert118 );
    
    #find last line
    my $ind = $#lines;
    $ind--;
    

    splice @lines, $ind , 0, @insert;
    _save_config( $filter, @lines );    
}


sub _remove_datatype118
{	
	my ( $lines,  $filter ) = @_;
    if ( grep {/Installed/} @$lines )
    {
        VERBOSE( "Datatype 118 is already updated" );
        return;
    }

    my $ind = 0;
    ++$ind while( @$lines[$ind] !~ /Autodesk.software.inventory/ );
	++$ind while( @$lines[$ind] !~ /Domain/ );
    splice ( @$lines, $ind , 1) ;
	++$ind while( @$lines[$ind] !~ /Operating.system/ );
	@$lines[$ind] = "   Operating^system,\n";
	++$ind;
	splice ( @$lines, $ind , 0, "   Location,\n   Installed)\n") ;
	_save_config( $filter, @$lines );    
	
}

sub _save_config
{
    
    my ( $cfg, @lines ) = @_; 
    &create_backup($cfg);
    open(OUT, ">$cfg") || die "Unable to overwrite filter file '$cfg'";
    print OUT @lines;
    close(OUT);
}

sub create_backup 
{
    my ($filter) = @_;
    DEBUG ( "Creating $filter backup " );
    my $new = "$filter.new";
    my $old = "$filter.old";

    open(FILE, $filter) || die "Unable to read filter file '$filter'";
    my @lines = <FILE>;
    close(FILE);
    
    open(OUT, ">$old") || die "Unable to create backup file '$old'";
    print OUT @lines;
    close(OUT);
}

sub get_configuration_dir()
{
    my $configdir;
    if ($^O eq 'MSWin32' )
    {
        my $cmd = `openit_freemould --script installation_info.oms`;
        chomp $cmd;
    
        $configdir =$cmd . "Configuration"
    }
    else
    {
        $configdir = `. /tmp/openit-cfg-shenv && echo \$CONFIG_DIR`;
    }
    chomp $configdir;
    DEBUG ("Got Confidir '$configdir'");
    return $configdir;
}


sub DEBUG 
{
    my ($log) = @_;
    $installationlog .= $log;
    if ( $opt_debug )
    {
        print ( "DEBUG: $log\n" );
    }
}

sub VERBOSE 
{
    my ($log) = @_;
    $installationlog .= $log;
    if ( $opt_verbose )
    {
        print ( "-- $log\n" );
    }
}

sub NOTICE 
{
    my ($log) = @_;
    $installationlog .= $log;
    print ( "NOTICE: $log\n" );
    
}

sub _grep_and_replace
{
    my ( $lines,  $grep_string, $new_value ) = @_;
    s/$grep_string/$new_value/g foreach(@$lines);
}


